// ╭──────────────────────────────────────────────────────────────────────────╮
// │                                                                          │
// │             Distributed under the CC0 public domain license.             │
// │   By Alison G. Watson. Attribution is encouraged, though not required.   │
// │                See licenses/cc0.txt for more information.                │
// │                                                                          │
// ├──────────────────────────────────────────────────────────────────────────┤
// │                                                                          │
// │ Dialogue compiler assembly statements.                                   │
// │                                                                          │
// ╰──────────────────────────────────────────────────────────────────────────╯

#include "d_compile.h"

static
i32 CodeABS(struct compiler *d, cstr reg) {
   struct token *tok = d->tb.get();
   i32 n;

   switch(tok->type) {
   case tok_number:
      n = faststrtoi32(tok->textV);
      if(n <= 0xFFFF) {
         if(!reg) {
            return n;
         } else if(d->tb.drop(tok_comma)) {
            tok = d->tb.get();

            if(TokIsKw(tok, reg)) return n;

            d->tb.unget();
         }
      }
      break;
   case tok_identi:
      n = -1;
      switch(Dlg_ItemName(tok->textV)) {
      case _dlg_item_page:
         tok = d->tb.get();
         if(tok->type == tok_identi) {
            n = Dlg_CheckNamePool(d, _name_pool_pages, tok->textV);
            if(n >= DPAGE_MAX || n < 0) {
               d->tb.err(&d->res, "bad page name '%s'", tok->textV);
               unwrap(&d->res);
            }
         } else {
            d->tb.unget();
         }
         break;
      case _dlg_item_failure:    n = DPAGE_FAILURE;    break;
      case _dlg_item_finished:   n = DPAGE_FINISHED;   break;
      case _dlg_item_unfinished: n = DPAGE_UNFINISHED; break;
      }
      if(n >= 0) {
         return PRG_BEG + d->def.pages[n];
      }
      break;
   case tok_mod:
      tok = d->tb.expc(&d->res, d->tb.get(), tok_identi, 0);
      unwrap(&d->res);

      n = Dlg_CheckNamePool(d, _name_pool_variables, tok->textV);
      if(n >= 0) {
         return VAR_END - n;
      } else {
         d->tb.err(&d->res, "unknown variable '%s'", tok->textV);
         unwrap(&d->res);
      }
      break;
   }

   d->tb.unget();

   return -1;
}

static
i32 CodeZPG(struct compiler *d, cstr reg) {
   struct token *tok = d->tb.get();

   if(tok->type == tok_number) {
      i32 n = faststrtoi32(tok->textV);

      if(n <= 0xFF) {
         if(!reg) {
            return n;
         } else if(d->tb.drop(tok_comma)) {
            tok = d->tb.get();

            if(TokIsKw(tok, reg)) return n;

            d->tb.unget();
         }
      }
   }

   d->tb.unget();

   return -1;
}

static
bool CodeAI(struct compiler *d, u32 code) {
   ifw(i32 n = CodeABS(d, nil), n >= 0) {
      Dlg_PushB1(d, code); unwrap(&d->res);
      Dlg_PushB2(d, n); unwrap(&d->res);
      return true;
   }

   return false;
}

static
bool CodeAX(struct compiler *d, u32 code) {
   ifw(i32 n = CodeABS(d, "X"), n >= 0) {
      Dlg_PushB1(d, code); unwrap(&d->res);
      Dlg_PushB2(d, n); unwrap(&d->res);
      return true;
   }

   return false;
}

static
bool CodeAY(struct compiler *d, u32 code) {
   ifw(i32 n = CodeABS(d, "Y"), n >= 0) {
      Dlg_PushB1(d, code); unwrap(&d->res);
      Dlg_PushB2(d, n); unwrap(&d->res);
      return true;
   }

   return false;
}

static
bool CodeII(struct compiler *d, u32 code) {
   if(d->tb.drop(tok_pareno)) {
      struct token *tok = d->tb.get();

      if(tok->type == tok_number) {
         i32 n = faststrtoi32(tok->textV);

         if(n <= 0xFFFF && d->tb.drop(tok_parenc)) {
            Dlg_PushB1(d, code); unwrap(&d->res);
            Dlg_PushB2(d, n); unwrap(&d->res);
            return true;
         }
      }

      d->tb.unget();
   }

   return false;
}

static
bool CodeIX(struct compiler *d, u32 code) {
   if(d->tb.drop(tok_pareno)) {
      struct token *tok = d->tb.get();

      if(tok->type == tok_number) {
         i32 n = faststrtoi32(tok->textV);

         if(n <= 0xFF && d->tb.drop(tok_comma)) {
            tok = d->tb.get();

            if(TokIsKw(tok, "X") && d->tb.drop(tok_parenc)) {
               Dlg_PushB1(d, code); unwrap(&d->res);
               Dlg_PushB1(d, n); unwrap(&d->res);
               return true;
            }

            d->tb.unget();
         }
      }

      d->tb.unget();
   }

   return false;
}

static
bool CodeIY(struct compiler *d, u32 code) {
   if(d->tb.drop(tok_pareno)) {
      struct token *tok = d->tb.get();

      if(tok->type == tok_number) {
         i32 n = faststrtoi32(tok->textV);

         if(n <= 0xFF && d->tb.drop(tok_parenc) && d->tb.drop(tok_comma)) {
            tok = d->tb.get();

            if(TokIsKw(tok, "Y")) {
               Dlg_PushB1(d, code); unwrap(&d->res);
               Dlg_PushB1(d, n); unwrap(&d->res);
               return true;
            }

            d->tb.unget();
         }
      }

      d->tb.unget();
   }

   return false;
}

static
bool CodeNP(struct compiler *d, u32 code) {
   Dlg_PushB1(d, code); unwrap(&d->res);
   return true;
}

#define CodeRI CodeZI

static
bool CodeVI(struct compiler *d, u32 code) {
   if(d->tb.drop(tok_hash)) {
      struct token *tok = d->tb.get();

      if(tok->type == tok_number) {
         i32 n = faststrtoi32(tok->textV);

         if(n <= 0xFF) {
            Dlg_PushB1(d, code); unwrap(&d->res);
            Dlg_PushB1(d, n); unwrap(&d->res);
            return true;
         }
      }

      d->tb.unget();
   }

   return false;
}

static
bool CodeZI(struct compiler *d, u32 code) {
   ifw(i32 n = CodeZPG(d, nil), n >= 0) {
      Dlg_PushB1(d, code); unwrap(&d->res);
      Dlg_PushB1(d, n); unwrap(&d->res);
      return true;
   }

   return false;
}

static
bool CodeZX(struct compiler *d, u32 code) {
   ifw(i32 n = CodeZPG(d, "X"), n >= 0) {
      Dlg_PushB1(d, code); unwrap(&d->res);
      Dlg_PushB1(d, n); unwrap(&d->res);
      return true;
   }

   return false;
}

static
bool CodeZY(struct compiler *d, u32 code) {
   ifw(i32 n = CodeZPG(d, "Y"), n >= 0) {
      Dlg_PushB1(d, code); unwrap(&d->res);
      Dlg_PushB1(d, n); unwrap(&d->res);
      return true;
   }

   return false;
}

void Dlg_GetStmt_Asm(struct compiler *d) {
   struct token *tok;

   #define DCD(n, op, ty) \
      tok = d->tb.reget(); \
      if(faststrcasechk(tok->textV, #op)) { \
         bool ret = Code##ty(d, DCD_##op##_##ty); \
         unwrap(&d->res); \
         if(ret) { \
            tok = d->tb.expc(&d->res, d->tb.get(), tok_semico, 0); \
            unwrap(&d->res); \
            return; \
         } \
      }
   #include "d_vm.h"

   tok = d->tb.reget();
   d->tb.err(&d->res, "%s no function found with this syntax", TokPrint(tok));
   unwrap(&d->res);
}

/* EOF */
